#include <stdlib.h>
#include <math.h>
#include "vox.h"

/* Set this to 0 for bigger (1.5 times) voxels */
#define	THIN_VOXELS	1

/***************************************************************
 * Some math
 */
#include "vmath.h"

/***************************************************************
 * Memory management for voxels.
 */

typedef	struct	_v_voxel	{
   struct _v_voxel *next;	/* double linked list */
   scrbox scrc;	/* Inclusive box coords */
   double z;	/* forward distance */
   transpinfo transp;
   colorinfo color;
} v_voxel;
typedef	struct	_tempct	{	/* temp color/transparency information */
   long r,g,b;	/* color sums */
   long t;	/* current transparency */
} tempct;

#define	VOX_PER_NODE	1000
typedef	struct	_v_voxarr	{
   struct _v_voxarr *next;
   v_voxel voxels[VOX_PER_NODE];
} v_voxarr;

static v_voxarr *arrays=0,*freearr=0;
static v_voxel *freevox=0;
static void done_mem(void) {
   v_voxarr *p=arrays,*n;
   while(p) {
      n=p->next; free(p); p=n;
   }
   freearr=arrays=0; freevox=0;
} /* done_mem */
static void init_mem(void) {
   done_mem();
   freearr=arrays=0;
   freevox=0;
} /* init_mem */
static void new_arr(void) {
   v_voxarr *p;
   int i;
   p=freearr;	/* Reuse already allocated */
   if(!p) {
      p=(v_voxarr*)malloc(sizeof(*p));
   } else {
      freearr=p->next;
   }
   if(!p) return;
   p->next=arrays; arrays=p;
   for(i=VOX_PER_NODE-1;i>=0;i--) {
      v_voxel *v=&(p->voxels[i]);
      v->next=freevox; freevox=v;	/* single linked list */
   }
} /* new_arr */
static v_voxel* new_vox(void) {
   v_voxel*v;
   if(!freevox) new_arr();
   if(!freevox) return 0;
   v=freevox; freevox=v->next;
   v->next=0;
   return v;
} /* v_voxel */
static void free_vox(v_voxel *p) {
   p->next=freevox; freevox=p;
} /* free_vox */

/***************************************************************
 * Generic (de)initialization code
 */
static int maxx,maxy,cx,cy;
static double z0,zclip;	/* Focus dist */
static v_voxel *head,*tail;
static pix *drawrow;
static tempct *tdrawrow;
double _vox_z_one=1.0;
int vox_init(int fov_ang,int _maxx,int _maxy) {
   if(drawrow) { free(drawrow); drawrow=0;}
   init_mem();
   maxx=_maxx; maxy=_maxy;
   drawrow=(pix*)malloc(sizeof(*drawrow)*maxx);
   if(!drawrow) return 0;
   tdrawrow=(tempct*)malloc(sizeof(*tdrawrow)*maxx);
   if(!tdrawrow) return 0;
   cx=maxx/2; cy=maxy/2;	/* Shift coords */
   z0=cx/tan(fov_ang/2.0*M_PI/180.0);
   _vox_z_one=z0;
   zclip=z0/1024;
   return 1;
} /* vox_init */
void vox_done(void) {
   done_mem();
   if(drawrow) free(drawrow); drawrow=0;
   if(tdrawrow) free(tdrawrow); drawrow=0;
} /* vox_done */

/***************************************************************
 * Voxel lists
 */
static dvec3 position,	/* point */
             forward,	/* View direction */
             up,	/* Y coord (invert for real screen) */
             right;	/* X coord */
static void init_dirs(void) {
   NORMALIZE3(forward);
   CROSS3(up,forward,right);
   NORMALIZE3(right);
   CROSS3(forward,right,up);	/* up is */
//   SCALE3(up,-1);	/* Now Y will be */
//   SCALE3(right,-1);	/* Now Y will be */
/* Now
   z=SCALAR3(c,forward);
   x=SCALAR3(c,right);
   y=SCALAR3(c,up);	for any 3d coordinate c.
 */
} /* init_dirs */
static void init_arrays(void) {
   v_voxarr*p,*n;
   p=arrays;
   while(p) {
      n=p->next; p->next=freearr; freearr=p; p=n;
   }
   arrays=0; freevox=0;
} /* init_arrays */
void vox_view(dvec3 *viewp,dvec3 *_forward,dvec3 *_up) {
   int i;
   for(i=0;i<N_DIMS;i++) {
      position[i]=(*viewp)[i];
      forward[i]=(*_forward)[i]; up[i]=(*_up)[i];
   }
   init_dirs();
   init_arrays();
} /* vox_view */
void vox_reset(void) {
   init_arrays();	/* Reset arrays list */
   head=0;	/* reset voxels list to empty */
 /* nothing? */
} /* vox_reset */

#define	transvec(x,y,z,vec) ((x)*(vec)[0]+(y)*(vec)[1]+(z)*(vec)[2])
#define	transx(x,y,z) transvec(x,y,z,right)
#define	transy(x,y,z) transvec(x,y,z,up)
#define	transz(x,y,z) transvec(x,y,z,forward)
int  vox_add(voxinfo*vi) {
   v_voxel*p;
   if(vi->transp>=TRANSPFULL) return 1;
   p=new_vox();
   if(!p) return 0;
   p->next=head;
   head=p;
   p->color=vi->color;
   p->transp=vi->transp;
   p->scrc=vi->scrc; p->z=vi->z;
   return VA_SUCCESS;
} /* vox_add */
#define	_invisible {v->invisible=1; v->transp=TRANSPFULL; return;}
void vox_check(voxinfo*v,int checksplit) {
   double hsize=v->size/2.0,qsize=hsize/1.0;
   double xx=(v->coord[X_DIM]+hsize-position[X_DIM]),	/* Center coords */
	  yy=(v->coord[Y_DIM]+hsize-position[Y_DIM]),
          zz=(v->coord[Z_DIM]+hsize-position[Z_DIM]);
   double z=transz(xx,yy,zz),iz;
   double slen;
   double x,y,lx,rx,uy,dy;
   v->split=0; v->invisible=0;
   if(checksplit) v->split=fabs(z) < hsize*2;	/* Should it to be splitted? */
   if(z<=zclip) {
       _invisible;	/* Nothing to do */
   }
   iz=z0/z;
   slen=(qsize+0.5)*iz;
   x=(transx(xx,yy,zz))*iz+maxx/2.0; lx=x-slen;
   if(lx>=maxx) _invisible;
   rx=x+slen;
   if(rx<0) _invisible;
   if(checksplit) {
      v->split=(lx<0) || (rx>=maxx);
   }
   y=transy(xx,yy,zz)*iz;
   y=y+maxy/2.0; uy=y-slen;
   if(uy>maxy) _invisible;
   dy=y+slen; if(dy<0) _invisible;
   if(checksplit) {
      v->split=(uy<0) || (dy>maxx);
   }
#if	THIN_VOXELS
   { double a=slen*0.6;
      lx=x-a;  rx=x+a; uy=y-a; dy=y+a;
      if(lx>=maxx || rx<0 || uy>=maxy || dy<0) { lx=rx=x; uy=y=y;}
   }
#endif
   if(lx<0) lx=0; if(rx>=maxx) rx=maxx-1;
   if(uy<0) uy=0; if(dy>=maxy) dy=maxy-1;
   v->scrc.left=lx; v->scrc.right=rx;
   v->scrc.up=uy; v->scrc.down=dy;
   v->z=z;
   v->scrsize=slen*2;
} /* vox_check */

/***************************************************************
 * Drawing code.
 */
int cmpvoxels(v_voxel *left,v_voxel*right) {
   if(left->z < right->z) return 1;	/* Z-order first */
   if(left->z > right->z) return -1;
   if(left->scrc.up < right->scrc.up) return 1;	/* lines */
   if(left->scrc.up > right->scrc.up) return -1;
   if(left->scrc.left < right->scrc.left) return 1;	/* columns */
   if(left->scrc.left > right->scrc.left) return -1;
   return 0;	/* equal? Impossible! ;) */
} /* cmpvoxel */
v_voxel *l_sort(v_voxel *l) {	/* Merge list sort */
   v_voxel*left,*right,*res,*rtail,*curr,*p;
   int flag;
   if(!l) return l;	/* Empty list */
   if(!l->next) return l;	/* list with only one element */
   left=right=0;
   for(flag=0;l;flag^=1) {
      p=l; l=l->next;
      if(flag) { p->next=left; left=p; }
      else     { p->next=right; right=p; }
   }
   left=l_sort(left); right=l_sort(right);
   res=0; rtail=0;
   while(left && right) {
      if(cmpvoxels(left,right)>0) {
	 curr=left; left=left->next;
      } else {
	 curr=right; right=right->next;
      }
      curr->next=0;
      if(!res) res=curr;
      if(!rtail) { rtail=curr;}
      else { rtail->next=curr; rtail=curr;}
   }
   if(left) {
      rtail->next=left;
      if(right && rtail) {
	 while(rtail->next) rtail=rtail->next;
      }
   }
   if(right) { rtail->next=right;}
   return res;
} /* l_sort */
static void sort_voxels(void) {
   head=l_sort(head);
} /* sort_voxels */
#define	BITSREMAIN	(TRANSPBITS-COLORCOMPBITS)
static void done_draw(int y) {
   long tr,tg,tb,tt;
   int x;
   for(x=0;x<maxx;x++) {
      tt=tdrawrow[x].t;
      tr=tdrawrow[x].r; tg=tdrawrow[x].g; tb=tdrawrow[x].b;
#if	0
      if(tt>TRANSPNONE) {
         tt=TRANSPONE-tt;
         tr=(tr*tt)>>TRANSPBITS;
         tg=(tg*tt)>>TRANSPBITS;
         tb=(tb*tt)>>TRANSPBITS;
      }
#endif
      tr>>=BITSREMAIN; if(tr>MAXSTOREDCOLORCOMP) tr=MAXSTOREDCOLORCOMP;
      tg>>=BITSREMAIN; if(tg>MAXSTOREDCOLORCOMP) tg=MAXSTOREDCOLORCOMP;
      tb>>=BITSREMAIN; if(tb>MAXSTOREDCOLORCOMP) tb=MAXSTOREDCOLORCOMP;
      drawrow[x].r=tr; drawrow[x].g=tg; drawrow[x].b=tb;
   }
   gr_pack(0,y,drawrow,maxx);
} /* done_draw */
static v_voxel*striplist(int y,v_voxel*list) {
   v_voxel*h=0,*t=0,*n;
   while(list) {
      n=list->next;
      if(list->scrc.down>y) {	/* still have to remain? */
	 if(!h)
	    h=list;
	 else
	    t->next=list;
         t=list;
      } else {	/* no */
	 free_vox(list);	/* return it to heap */
      }
      list=n;
   }
   if(t) t->next=0;
   return h;
} /* striplist */
static v_voxel* draw_row(int y,v_voxel*list) {
   int x,xe;
   long tr,tg,tb,it,t,tt;
   v_voxel*p=list;
   for(x=0;x<maxx;x++) {
      tdrawrow[x].r=tdrawrow[x].g=tdrawrow[x].b=0; tdrawrow[x].t=TRANSPFULL;
   }
   while(p) {
      tr=p->color.r; tg=p->color.g; tb=p->color.b;
      t=p->transp;
      it=TRANSPFULL-t;
      for(x=p->scrc.left,xe=p->scrc.right;x<=xe;x++) {
//if(y==100 && x==160)__emit__(0xcc);
         tt=tdrawrow[x].t; if(tt<=TRANSPNONE) continue;
         tt=(tt*it)+TRANSPFULL*3/4; tt>>=TRANSPBITS;
         tdrawrow[x].r+=tr*tt; tdrawrow[x].g+=tg*tt; tdrawrow[x].b+=tb*tt;
         tdrawrow[x].t=(tdrawrow[x].t*t+TRANSPFULL/2)>>TRANSPBITS;
      }
      p=p->next;
   }
   done_draw(y);
   return striplist(y,list);
} /* draw_row */
static v_voxel* merge_lists(v_voxel*a,v_voxel*b) {
   v_voxel*l,*t,*p;
   l=0;
   if(!a && !b) return l;
   while(a && b) {
      if(a->z<=b->z) {
         p=a; a=a->next;
      } else {
         p=b; b=b->next;
      }
      if(l)
         t->next=p;
      else
         l=p;
      t=p;
   }
   p=a?a:b;
   if(l) t->next=p; else l=p;
   return l;
} /* merge_lists */
v_voxel* split_list(v_voxel**list,int below) {
   v_voxel*cl,*l,*r,*p,*lt,*rt,*n;
   l=lt=r=rt=0;
   cl=*list;
   while(cl) {
      p=cl; n=p->next;
      if(p->scrc.up>=below) {
         if(lt)
            lt->next=p;
         else
            l=p;
         lt=p;
      } else {
         if(rt) rt->next=p;
         else r=p;
         rt=p;
      }
      cl=n;
   }
   if(lt) lt->next=0; if(rt) rt->next=0;
   *list=l;
   return r;
} /* split_list */
v_voxel* div_vert(int notb,int below,v_voxel*list) {
   int l;
   v_voxel *rem1,*rem2,*tail1,*tail2;
   l=below-notb;
   if(l>1) {
      l/=2;
      rem1=split_list(&list,notb+l);
      rem1=div_vert(notb,notb+l,rem1);
      rem1=merge_lists(rem1,list);
      rem1=div_vert(notb+l,below,rem1);
      return rem1;
   } else {
      rem1=draw_row(notb,list);
      return rem1;
   }
} /* div_vert */
void vox_draw(void) {
   int x,y;
   sort_voxels();
   if(div_vert(0,maxy,head)) printf("Error in vox_draw!\n");
} /* vox_draw */
